#pragma once
#include "dip_plane.h"
using namespace dip;

#include<cmath>
#include<cassert>
#include<vector>
using namespace std;

namespace dip{

template<class T>
class Image : public plane<T>{
public:
	Image();
	Image(int width, int height);
	Image(const Image<T> & rh);

	// tell image to slice itselft into blocks ( grid )
	void SliceToBlocks(int block_width, int block_height);
	bool getBlock(Block<T> & b, int block_x, int block_y, int block_off_x, int block_off_y);

	Image<T> ConvSlow(const Image<float> & filter);
	Image<T> ConvX(const vector<float> & filter);
	Image<T> ConvY(const vector<float> & filter);
	Image<T> ConvFast(const vector<float> & filter);
	Image<T> ConvFast(const vector<float> & filter, plane<bool> & mask);

	Image<T> DownSample(int factor);
	Image<T> UpSample(int factor);
	Image<T> Dx();
	Image<T> Dy();
	Image<T> Dx(plane<bool> & mask_in, plane<bool> & mask_out);
	Image<T> Dy(plane<bool> & mask_in, plane<bool> & mask_out);


	// A -- B
	// |    |
	// C -- D
	inline T lerp_entry(double x, double y) const{
		int xf = (int)floor(x);
		int yf = (int)floor(y);
		int xc = xf + 1;
		int yc = yf + 1;

		if(x >= m_width - 1){
			xf = m_width - 1;
			xc = m_width - 1;
		}

		if(y >= m_height - 1){
			yf = m_height - 1;
			yc = m_height - 1;
		}

		double dx = x - xf;
		double dy = y - yf;

		T A_B = (T)(entry(xf, yf)*(1 - dx)  + entry(xc, yf)*dx);
		T C_D = (T)(entry(xf, yc)*(1 - dx)  + entry(xc, yc)*dx);
		T out = (T)(A_B*(1 - dy) + C_D*dy);
		return out;
	}

	Image<T> & operator=(const Image<T> & rh);
};

template<class T>
struct RGBpel{
	RGBpel():R(0), G(0), B(0){ }
	RGBpel(T iR, T iG, T iB):R(iR), G(iG), B(iB){}
	RGBpel(T rh):R(rh), G(rh), B(rh){}

	T B;
	T G;
	T R;

	inline T getIntensity(){return  static_cast<T>(0.299*R + 0.587*G + 0.114*B)};
	inline RGBpel operator+(const RGBpel & rh)const{ return RGBpel(R + rh.R, G + rh.G, B + rh.B); }
	inline RGBpel operator-(const RGBpel & rh)const{ return RGBpel(R - rh.R, G - rh.G, B - rh.B); }
	inline RGBpel operator*(const RGBpel & rh)const{ return RGBpel(R * rh.R, G * rh.G, B * rh.B); }
	inline RGBpel operator/(const RGBpel & rh)const{ return RGBpel(R / rh.R, G / rh.G, B / rh.B); }
	inline RGBpel operator*(const double & rh)const{ return RGBpel(static_cast<T>(R * rh), static_cast<T>(G * rh), static_cast<T>(B * rh)); }
	inline RGBpel operator/(const double & rh)const{ return RGBpel(static_cast<T>(R / rh), static_cast<T>(G / rh), static_cast<T>(B / rh)); }
	inline RGBpel & operator=(int rh){B = 0;G = 0;R = 0;return *this;}
	inline RGBpel & operator+=(const RGBpel & rh){ R += rh.R; G += rh.G; B += rh.B; return *this;}
};

/*******************************************************************************************
implementation of Image
*******************************************************************************************/
template<class T>
Image<T>::Image() : plane<T>(){
}

template<class T>
Image<T>::Image(int width, int height) : plane<T>(width, height){
}

template<class T>
Image<T>::Image(const dip::Image<T> & rh) : plane<T>(rh){
}

template<class T>
Image<T> & Image<T>::operator=(const Image<T> & rh){
	resize(rh.m_width, rh.m_height);
	memmove((void*)m_buf, (void*)rh.m_buf, m_width*m_height*sizeof(T));
	return *this;
}


template<class T>
Image<T> Image<T>::ConvSlow(const Image<float> & filter){
	Image<T> result(m_width, m_height);

	for (int y = 0; y < result.m_height; y++) 
	{
		for (int x = 0; x < result.m_width; x++) 
		{
			float m = 0.0f;
			T v = 0;

			for (int ky = 0; ky < filter.m_height; ky++) 
			{
				for (int kx = 0; kx < filter.m_width; kx++)
				{
					int	cx = x + kx - filter.m_width / 2;
					int	cy = y + ky - filter.m_height / 2;

					if (!((cx < 0) || (cy < 0) || (cx >= result.m_width) || (cy >= result.m_height))) 
					{
						m += filter.entry(kx, ky);
						v +=  entry(cx,cy)*filter.entry(kx, ky);
					}
				}
			}

			result.entry(x, y) = v / m;
		}
	}
	return result;
}

template<class T>
Image<T> Image<T>::ConvX(const vector<float> & filter){
	Image<T> result(m_width, m_height);
	int filter_size = (int)filter.size();

	float* filter_array = new float[filter_size];
	for(int i = 0; i < filter_size; i++)
		filter_array[i] = filter[i];

	for(int y = 0; y < m_height; y++){
		for(int x = 0; x < m_width; x++){
			float m = 0.0f;
			T v = 0;
			for (int kx = 0; kx < filter_size; kx++){
				int cx = x + kx - filter_size/ 2;
				if (!((cx < 0) || (cx >= result.m_width))){
					m += filter_array[kx];
					v += entry(cx, y)*filter_array[kx];
				}
			}

			result.entry(x, y) = v / m;		
		}
	}
	return result;
}

template<class T>
Image<T> Image<T>::ConvY(const vector<float> & filter){
	Image<T> result(m_width, m_height);
	int filter_size = (int)filter.size();

	float* filter_array = new float[filter_size];
	for(int i = 0; i < filter_size; i++)
		filter_array[i] = filter[i];

	for(int x = 0; x < m_width; x++){
		for(int y = 0; y < m_height; y++){
			float m = 0.0f;
			T v = 0;
			for(int ky = 0; ky < filter_size;  ky++){
				int	cy = y + ky - filter_size / 2;
				if (!((cy < 0) || (cy >= result.m_height))) {
					m += filter_array[ky];
					v +=  entry(cx,cy)*filter_array[ky];
				}
			}

			result.entry(x, y) = v / m;	
		}
	}
}

template<class T>
Image<T> Image<T>::ConvFast(const vector<float> & filter){
	Image<T> stage1(m_width, m_height);
	Image<T> stage2(m_width, m_height);

	int filter_size = (int)filter.size();
	float* filter_array = new float[filter_size];
	for(int i = 0; i < filter_size; i++)
		filter_array[i] = filter[i];

	// stage 1;
	for(int y = 0; y < m_height; y++){
		for(int x = 0; x < m_width; x++){
			float m = 0.0f;
			T v = 0;
			for (int kx = 0; kx < filter_size; kx++){
				int cx = x + kx - filter_size/ 2;
				if (!((cx < 0) || (cx >= m_width))){
					m += filter_array[kx];
					v += entry(cx, y)*filter_array[kx];
				}
			}

			stage1.entry(x, y) = v / m;		
		}
	}

	// stage 2;
	for(int x = 0; x < m_width; x++){
		for(int y = 0; y < m_height; y++){
			float m = 0.0f;
			T v = 0;
			for(int ky = 0; ky < filter_size;  ky++){
				int	cy = y + ky - filter_size / 2;
				if (!((cy < 0) || (cy >= m_height))) {
					m += filter_array[ky];
					v +=  stage1.entry(x,cy)*filter_array[ky];
				}
			}

			stage2.entry(x, y) = v / m;	
		}
	}

	return stage2;
}


template<class T>
Image<T> Image<T>::ConvFast(const vector<float> & filter, plane<bool> & mask){
	Image<T> stage1(m_width, m_height);
	Image<T> stage2(m_width, m_height);

	int filter_size = (int)filter.size();
	float* filter_array = new float[filter_size];
	for(int i = 0; i < filter_size; i++)
		filter_array[i] = filter[i];

	// stage 1;
	for(int y = 0; y < m_height; y++){
		for(int x = 0; x < m_width; x++){

			if( !mask.entry(x, y) )
				continue;


			float m = 0.0f;
			T v = 0;
			for (int kx = 0; kx < filter_size; kx++){
				int cx = x + kx - filter_size/ 2;
				if (!((cx < 0) || (cx >= m_width)) && mask.entry(x,y)){
					m += filter_array[kx];
					v += entry(cx, y)*filter_array[kx];
				}
			}

			stage1.entry(x, y) = v / m;		
		}
	}

	// stage 2;
	for(int x = 0; x < m_width; x++){
		for(int y = 0; y < m_height; y++){

			if( !mask.entry(x, y) )
				continue;

			float m = 0.0f;
			T v = 0;
			for(int ky = 0; ky < filter_size;  ky++){
				int	cy = y + ky - filter_size / 2;
				if (!((cy < 0) || (cy >= m_height)) && mask.entry(x,y)) {
					m += filter_array[ky];
					v +=  stage1.entry(x,cy)*filter_array[ky];
				}
			}

			stage2.entry(x, y) = v / m;	
		}
	}

	return stage2;
}

template<class T>
Image<T> Image<T>::DownSample(int factor){
	Image<T> result(m_width/factor, m_height/factor);
	for(int i = 0; i < result.m_width; i++){
		for(int j = 0; j < result.m_height; j++){
			result.entry(i,j) = this->entry(factor*i, factor*j);
		}
	}
	return result;
}

template<class T>
Image<T> Image<T>::UpSample(int factor){
	Image<T> result(m_width*factor, m_height*factor);
	for(int j = 0; j < result.m_height; j++){
		for(int i = 0; i < result.m_width; i++){
			result.entry(i,j) = lerp_entry(i/(double)factor, j/(double)factor);
		}
	}
	return result;
}

template<class T>
Image<T> Image<T>::Dx(){
	Image<T> result(m_width, m_height);
	for(int j = 1; j < m_height - 1; j++){
		for(int i = 1; i < m_width - 1; i++){
			result.entry(i,j) = (entry(i+1, j) - entry(i-1, j))/2;
		}
	}
	return result;
}

template<class T>
Image<T> Image<T>::Dy(){
	Image<T> result(m_width, m_height);
	for(int j = 1; j < m_height - 1; j++){
		for(int i = 1; i < m_width - 1; i++){
			result.entry(i,j) = (entry(i, j+1) - entry(i, j-1))/2;
		}
	}
	return result;
}

template<class T>
Image<T> Image<T>::Dx(plane<bool> & mask_in, plane<bool> & mask_out){
	Image<T> result(m_width, m_height);
	mask_out.resize(m_width, m_height);
	for(int j = 1; j < m_height - 1; j++){
		for(int i = 1; i < m_width - 1; i++){
			if( mask_in.entry(i+1,j) && mask_in.entry(i- 1,j)){
				result.entry(i,j) = (entry(i+1, j) - entry(i-1, j))/2;
				mask_out.entry(i, j) = true;
			}else{
				mask_out.entry(i, j) = false;
			}
		}
	}
	return result;
}

template<class T>
Image<T> Image<T>::Dy(plane<bool> & mask_in, plane<bool> & mask_out){
	Image<T> result(m_width, m_height);
	for(int j = 1; j < m_height - 1; j++){
		for(int i = 1; i < m_width - 1; i++){
			if( mask_out.entry(i, j+1) && mask_out.entry(i, j-1) ){
				result.entry(i,j) = (entry(i, j+1) - entry(i, j-1))/2;
			}else{
				mask_out.entry(i, j) = false;
			}
		}
	}
	return result;
}

/*******************************************************************************************
typedefs
*******************************************************************************************/

typedef RGBpel<unsigned char> uRGB;
typedef RGBpel<float> fRGB;

typedef Image<unsigned char> GrayImage;
typedef Image<float> FloatImage;
typedef Image<uRGB>	RgbImage;
typedef Image<fRGB> FloatRgbImage;


} // end of namespace dip

