#pragma once
#include "dip_image.h"
#include "math_matrix.h"
#include "math_vector.h"
using namespace math;
using namespace dip;

#include <vector>
#include <cassert>
using namespace std;

namespace dip{
	namespace routine{
		template<class T>
		Image<T> AffineTransform(Image<T> & in, vector<double> param);

		template<class T>
		Image<T> PerspectiveTransform(Image<T> & in, vector<double> param);

		RgbImage ConvertToRgb(GrayImage & in);

		inline Matrix getAffineMatrix(vector<double> & param){
			assert(param.size() == 6);
			Matrix result(3, 3);
			result.entry(0, 0) = param[0];
			result.entry(0, 1) = param[1];
			result.entry(0, 2) = param[2];
			result.entry(1, 0) = param[3];
			result.entry(1, 1) = param[4];
			result.entry(1, 2) = param[5];
			result.entry(2, 2) = 1.0;
			return result;
		}

		inline Matrix getPerspectiveMatrix(vector<double> & param){
			assert(param.size() == 8);
			Matrix result(3, 3);
			result.entry(0,0) = param[0];
			result.entry(0,1) = param[1];
			result.entry(0,2) = param[2];
			result.entry(1,0) = param[3];
			result.entry(1,1) = param[4];
			result.entry(1,2) = param[5];
			result.entry(2,0) = param[6];
			result.entry(2,1) = param[7];
			result.entry(2,2) = 1.0;
			return result;
		}


		/*******************************************************
		implementation
		********************************************************/
		template<class T>
		Image<T> AffineTransform(Image<T> & in, vector<double> param){
			assert(param.size() == 6);
			Image<T> result(in.m_width, in.m_height);
			double a1 = param[0];
			double b1 = param[1];
			double c1 = param[2];
			double a2 = param[3];
			double b2 = param[4];
			double c2 = param[5];

			double k = 1/(a1*b2-b1*a2);
			for(int j = 0; j < in.m_height; j++){
				for(int i = 0; i < in.m_width; i++){
					double x, y;
					x =k*(  b2 *(i - c1) - b1*(j - c2));
					y =k*((-a2)*(i - c1) + a1*(j - c2));
					if( in.checkInside(x,y))
						result.entry(i, j) = in.lerp_entry(x, y);
					else
						result.entry(i, j) = 0;
				}
			}
			return result;
		}

		template<class T>
		Image<T> PerspectiveTransform(Image<T> & in, vector<double> param){
			Image<T> result(in.m_width, in.m_height);
			Matrix t = getPerspectiveMatrix(param).getInverse();

			for(int j = 0; j < in.m_height; j++){
				for(int i = 0; i < in.m_width; i++){
					Vector3d P(i, j, 1.0);
					Vector3d P_ = t*P;
					P_.x /= P_.z;
					P_.y /= P_.z;
					if( in.checkInside(P_.x, P_.y) )
						result.entry(i,j) = in.lerp_entry(P_.x,P_.y);
					else
						result.entry(i,j) = 0;
					
				}
			}
			return result;
		}

		template<class T>
		Image<T> WarpImage(Image<T> & in, Matrix & transform){
			Image<T> result(in.m_width, in.m_height);
			Matrix t = transform.getInverse();

			for(int j = 0; j < in.m_height; j++){
				for(int i = 0; i < in.m_width; i++){
					Vector3d P(i, j, 1.0);
					Vector3d P_ = t*P;
					P_.x /= P_.z;
					P_.y /= P_.z;
					if( in.checkInside(P_.x, P_.y) )
						result.entry(i,j) = in.lerp_entry(P_.x,P_.y);
					else
						result.entry(i,j) = 0;
					
				}
			}
			return result;
		
		}

		template<class T>
		Image<T> WarpImage(Image<T> & in, Matrix & transform, plane<bool> & mask){
			Image<T> result(in.m_width, in.m_height);
			mask.resize(in.m_width, in.m_height);
			Matrix t = transform.getInverse();

			for(int j = 0; j < in.m_height; j++){
				for(int i = 0; i < in.m_width; i++){
					Vector3d P(i, j, 1.0);
					Vector3d P_ = t*P;
					P_.x /= P_.z;
					P_.y /= P_.z;
					if( in.checkInside(P_.x, P_.y) ){
						result.entry(i,j) = in.lerp_entry(P_.x,P_.y);
						mask.entry(i,j) = true;
					}
					else{
						result.entry(i,j) = 0;
						mask.entry(i,j) = false;
					}
					
				}
			}
			return result;
		
		}


	}// end of io namespace
}// end of dip namespace


