#include "math_matrix.h"
using namespace math;

#include<cstring>
#include<cassert>
#include<exception>
using namespace std;

Matrix::Matrix(){
	m_buf = NULL;
	m_row = 0;
	m_col = 0;
}

Matrix::Matrix(int row, int col){
	assert(row > 0 && col > 0);
	m_buf = new double[row*col];
	memset(m_buf, 0, row*col*sizeof(double));
	m_row = row;
	m_col = col;
}

Matrix::Matrix(const Matrix & rh){
	if(rh.m_buf != NULL){
		m_row = rh.m_row;
		m_col = rh.m_col;
		m_buf = new double[m_row*m_col];
		memmove((void*)m_buf, (void*)rh.m_buf, m_row*m_col*sizeof(double));

	}
	else{
		m_buf = NULL;
		m_row = 0;
		m_col = 0;
	}
}

Matrix::~Matrix(){
	if( m_buf != NULL)
		delete [] m_buf;
}

bool Matrix::resize(int row, int col){
	if( m_row * m_col != row *col){
		reset();
		m_buf = new double[row*col];
	}
	m_row = row;
	m_col = col;
	memset(m_buf, 0, row*col*sizeof(double));
	return true;
}

void Matrix::reset(){
	if(m_buf != NULL)
		delete [] m_buf;
	m_buf = NULL;
	m_row = 0;
	m_col = 0;
}

void Matrix::identity(){
	assert(m_col != 0 && m_row == m_col);
	memset((void*)m_buf, 0, m_row*m_col);

	for(int i = 0; i < m_col; i++)
		entry(i, i) = 1.0;
}

void Matrix::zero(){
	if( m_col == 0 || m_row == 0 || m_buf == NULL )
		throw exception("Matrix::Zero: the matrix is empty");
	memset((void*)m_buf, 0, m_row*m_col);
}

Matrix Matrix::getTranspose(){
	assert(m_buf != NULL);
	Matrix result(m_col, m_row);
	for(int i = 0; i < m_row; i++){
		for(int j = 0; j < m_col; j++)
			result.entry(j, i) = entry(i, j);
	}
	return result;
}

Matrix Matrix::getInverse(){
	//if(m_row == 0 || m_row != m_col || m_row < 2)
	//	return false;
	assert( m_row > 1 && m_row == m_col);

	Matrix leftMx = *this;
	Matrix rightMx(m_row, m_col);
	rightMx.identity();

	//for every row as pivot row
	for(int pivot = 0; pivot < leftMx.m_row; pivot++){
		if(leftMx.entry(pivot, pivot) == 0){
			//if leading entry fo pivot row is 0, then handle this.

			//check whether this row is all-zero row
			bool zero_row = true;

			if(pivot != leftMx.m_col - 1){
				for(int i = pivot + 1; i < leftMx.m_col; i++){
					if(leftMx.entry(pivot, i) != 0){
						zero_row = false;
						break;
					}
				}
			}
			
			if(zero_row)
				throw exception("matrix inverse: rank is not enough!");

			//swap another row with non-zero pivot-col to this row.
			for(int i = pivot + 1;  i < m_row; i++){
				if(leftMx.entry(i,pivot) != 0.0){
					leftMx.AddRow(i, pivot);
					rightMx.AddRow(i, pivot);
					break;
				}
			}

			if(leftMx.entry(pivot, pivot) == 0)
				throw exception("matrix inverse: rank is not enough!");

		}

		//normalize the pivot row
		static double leading_ent = 0.0;
		leading_ent = leftMx.entry(pivot, pivot);  //save leading entry in another var to save time to compute idx
		for(int i = pivot; i < m_col; i++)
			leftMx.entry(pivot, i) /= leading_ent;
		for(int i = 0; i < leftMx.m_col; i++)
			rightMx.entry(pivot, i) /= leading_ent;

		for(int q = 0; q < leftMx.m_row; q++){
			if( q != pivot){
				static double scale = 0.0;
				scale = -leftMx.entry(q, pivot);
				
				for(int r = pivot; r < m_col; r++)
					leftMx.entry(q, r) +=  scale*leftMx.entry(pivot, r);
				for(int r = 0; r < rightMx.m_col; r++)
					rightMx.entry(q, r) += scale*rightMx.entry(pivot, r);
			}
		}
	}

	return rightMx;
}

void Matrix::SwapRow(int row1, int row2){
	if( row1 < 0 || row1 >= m_row || row2 < 0 || row2 >= m_row)
		throw exception("SwapRow: incorrect parameter");
	
	for(int col = 0; col < m_col; col++){
		double temp = entry(row1, col);
		entry(row1, col) = entry(row2, col);
		entry(row2, col) = temp;
	}
}

void Matrix::AddRow(int row1, int row2){
	if( row1 < 0 || row1 >= m_row || row2 < 0 || row2 >= m_row)
		throw exception("SwapRow: incorrect parameter");
	
	for(int col = 0; col < m_col; col++){
		entry(row2, col) += entry(row1, col);
	}
}


Matrix & Matrix::operator=(const Matrix & rh){
	resize(rh.m_row, rh.m_col);
	for(int i = 0; i < m_row*m_col; i++)
		m_buf[i] = rh.m_buf[i];
	return *this;
};

Matrix Matrix::operator*(const Matrix & rh) const{
	assert(rh.m_row == m_col && m_buf != NULL);
	Matrix result(m_row, rh.m_col);
	for(int i = 0; i < result.m_row; i++){
		for(int j = 0; j < result.m_col; j++){
			result.entry(i, j) = 0;
			for(int k = 0; k < m_row; k++){
				result.entry(i,j) += entry(i,k)*rh.entry(k,j);
			}
		}
	}
	return result;
}

Matrix Matrix::operator*(double scale) const{
	assert(m_buf != NULL);
	Matrix result(m_row, m_col);
	for(int i = 0; i < m_row*m_col; i++)
		result.m_buf[i] = m_buf[i]*scale;
	return result;
}

Matrix Matrix::operator+(const Matrix & rh) const{
	assert(m_row == rh.m_row && m_col == rh.m_col);
	Matrix result(m_row, m_col);
	for(int i = 0; i < m_row*m_col; i++)
		result.m_buf[i] = m_buf[i] + rh.m_buf[i];
	return result;
}


Matrix & Matrix::operator*=(const Matrix & rh){
	Matrix result = (*this)*rh;
	(*this) = rh;
	return *this;
}

Matrix & Matrix::operator*=(double scale){
	assert(m_buf != NULL);
	for(int i = 0; i < m_row*m_col; i++)
		m_buf[i] *= scale;
	return *this;

}
Matrix & Matrix::operator+=(const Matrix & rh){
	assert(m_row == rh.m_row && m_col == rh.m_col);
	for(int i = 0; i < m_row*m_col; i++)
		m_buf[i] += rh.m_buf[i];
	return *this;
}
