#include "rgbe.h"

/*
sample minimal reading code:
  
	f = fopen(image_filename, "rb");
	RGBE_ReadHeader(f, &image_width, &image_height,NULL);
	
	image = (float *)malloc(sizeof(float)*3*image_width*image_height);
	RGBE_ReadPixels_RLE(f,image,image_width,image_height);

	fclose(f);

 */

int rgbe_error(int rgbe_error_code, char *msg)
{
  switch (rgbe_error_code) {
  case rgbe_read_error:
    perror("RGBE read error");
    break;
  case rgbe_write_error:
    perror("RGBE write error");
    break;
  case rgbe_format_error:
    fprintf(stderr,"RGBE bad file format: %s\n",msg);
    break;
  default:
  case rgbe_memory_error:
    fprintf(stderr,"RGBE error: %s\n",msg);
  }
  return RGBE_RETURN_FAILURE;
}

static void rgbe2double(double *red, double *green, double *blue, unsigned char rgbe[4])
{
	double f;

	if (rgbe[3]) 
	{   /*nonzero pixel*/
		f = ldexp(1.0,rgbe[3]-(int)(128+8));
		*red = rgbe[0] * f;
		*green = rgbe[1] * f;
		*blue = rgbe[2] * f;
	}
	else
		*red = *green = *blue = 0.0;
}

static void rgbe2float(float*red, float *green, float *blue, unsigned char rgbe[4])
{
	double f;

	if (rgbe[3]) 
	{   /*nonzero pixel*/
		f = ldexp(1.0,rgbe[3]-(int)(128+8));
		*red = (float)(rgbe[0] * f);
		*green = (float)(rgbe[1] * f);
		*blue = (float)(rgbe[2] * f);
	}
	else
		*red = *green = *blue = 0.0f;
}

int RGBE_ReadHeader(FILE *fp, int *width, int *height, rgbe_header_info *info)
{
	char buf[128];
	int found_format;
	float tempf;
	int i;

	found_format = 0;
	if (info) 
	{
		info->valid = 0;
		info->programtype[0] = 0;
		info->gamma = info->exposure = 1.0;
	}
	if (fgets(buf,sizeof(buf)/sizeof(buf[0]),fp) == NULL)
		return rgbe_error(rgbe_read_error,NULL);

	// Read #? Radiance
	if ((buf[0] != '#')||(buf[1] != '?')) 
	{
		/* if you want to require the magic token then uncomment the next line */
		return rgbe_error(rgbe_format_error, "bad initial token");
	}
	else if (info) 
	{
		info->valid |= RGBE_VALID_PROGRAMTYPE;
		for(i=0;i<sizeof(info->programtype)-1;i++) 
		{
			if ((buf[i+2] == 0) || isspace(buf[i+2]))
				break;
			info->programtype[i] = buf[i+2];
		}
		info->programtype[i] = 0;
		if (fgets(buf,sizeof(buf)/sizeof(buf[0]),fp) == 0)
			return rgbe_error(rgbe_read_error,NULL);
	}
	for(;;) 
	{
		if ((buf[0] == 0)||(buf[0] == '\n'))
			return rgbe_error(rgbe_format_error,"no FORMAT specifier found");

		/* format found so break out of loop */
		else if (strcmp(buf,"FORMAT=32-bit_rle_rgbe\n") == 0)
			break;

		else if (info && (sscanf(buf,"GAMMA=%g",&tempf) == 1)) 
		{
			info->gamma = tempf;
			info->valid |= RGBE_VALID_GAMMA;
		}
		else if (info && (sscanf(buf,"EXPOSURE=%g",&tempf) == 1)) 
		{
			info->exposure = tempf;
			info->valid |= RGBE_VALID_EXPOSURE;
		}
		if (fgets(buf,sizeof(buf)/sizeof(buf[0]),fp) == 0)
			return rgbe_error(rgbe_read_error,NULL);
	}
	if (fgets(buf,sizeof(buf)/sizeof(buf[0]),fp) == 0)
		return rgbe_error(rgbe_read_error,NULL);
	
	// One empty line before size specification
	for(;;)
	{
		if (strcmp(buf,"\n") == 0)
			break;
		else
			if (fgets(buf,sizeof(buf)/sizeof(buf[0]),fp) == 0)
				return rgbe_error(rgbe_read_error,NULL);
//		return rgbe_error(rgbe_format_error, "missing blank line after FORMAT specifier");
	}
	
	// Get image size
	if (fgets(buf,sizeof(buf)/sizeof(buf[0]),fp) == 0)
		return rgbe_error(rgbe_read_error,NULL);
	if (sscanf(buf,"-Y %d +X %d",height,width) < 2)
		return rgbe_error(rgbe_format_error,"missing image size specifier");
	return RGBE_RETURN_SUCCESS;
}


/* simple read routine.  will not correctly handle run length encoding */
int RGBE_ReadPixels(FILE *fp, float *data, int numpixels)
{
	unsigned char rgbe[4];

	while(numpixels-- > 0) 
	{
		if (fread(rgbe, sizeof(rgbe), 1, fp) < 1)
			return rgbe_error(rgbe_read_error,NULL);
		rgbe2float(&data[RGBE_DATA_RED], &data[RGBE_DATA_GREEN], &data[RGBE_DATA_BLUE],rgbe);
		data += RGBE_DATA_SIZE;
	}
	return RGBE_RETURN_SUCCESS;
}

int RGBE_ReadPixels_RLE(FILE *fp, float *data, int scanline_width,	int num_scanlines)
{
	unsigned char rgbe[4], *scanline_buffer, *ptr, *ptr_end;
	int i, count;
	unsigned char buf[2];

	if ((scanline_width < 8)||(scanline_width > 0x7fff))
	/* run length encoding is not allowed so read flat*/
		return RGBE_ReadPixels(fp,data,scanline_width*num_scanlines);
	scanline_buffer = NULL;
	/* read in each successive scanline */
	while(num_scanlines > 0) 
	{
		if (fread(rgbe,sizeof(rgbe),1,fp) < 1)
		{
			free(scanline_buffer);
			return rgbe_error(rgbe_read_error,NULL);
		}
		if ((rgbe[0] != 2)||(rgbe[1] != 2)||(rgbe[2] & 0x80)) 
		{
			/* this file is not run length encoded */
			rgbe2float(&data[0],&data[1],&data[2],rgbe);
			data += RGBE_DATA_SIZE;
			free(scanline_buffer);
			return RGBE_ReadPixels(fp,data,scanline_width*num_scanlines-1);
		}
		if ((((int)rgbe[2])<<8 | rgbe[3]) != scanline_width) 
		{
			free(scanline_buffer);
			return rgbe_error(rgbe_format_error,"wrong scanline width");
		}
		if (scanline_buffer == NULL)
			scanline_buffer = (unsigned char *)
		malloc(sizeof(unsigned char)*4*scanline_width);
		if (scanline_buffer == NULL) 
			return rgbe_error(rgbe_memory_error,"unable to allocate buffer space");

		ptr = &scanline_buffer[0];
		/* read each of the four channels for the scanline into the buffer */
		for(i=0;i<4;i++) 
		{
			ptr_end = &scanline_buffer[(i+1)*scanline_width];
			while(ptr < ptr_end) 
			{
				if (fread(buf,sizeof(buf[0])*2,1,fp) < 1) 
				{
					free(scanline_buffer);
					return rgbe_error(rgbe_read_error,NULL);
				}
				if (buf[0] > 128) 
				{
					/* a run of the same value */
					count = buf[0]-128;
					if ((count == 0)||(count > ptr_end - ptr)) 
					{
						free(scanline_buffer);
						return rgbe_error(rgbe_format_error,"bad scanline data");
					}
				while(count-- > 0)
					*ptr++ = buf[1];
				}
				else 
				{
					/* a non-run */
					count = buf[0];
					if ((count == 0)||(count > ptr_end - ptr)) 
					{
						free(scanline_buffer);
						return rgbe_error(rgbe_format_error,"bad scanline data");
					}
					*ptr++ = buf[1];
					if (--count > 0) 
					{
						if (fread(ptr,sizeof(*ptr)*count,1,fp) < 1) 
						{
							free(scanline_buffer);
							return rgbe_error(rgbe_read_error,NULL);
						}
						ptr += count;
					}
				}
			}
		}
		/* now convert data from buffer into floats */
		for(i=0;i<scanline_width;i++) 
		{
		  rgbe[0] = scanline_buffer[i];
		  rgbe[1] = scanline_buffer[i+scanline_width];
		  rgbe[2] = scanline_buffer[i+2*scanline_width];
		  rgbe[3] = scanline_buffer[i+3*scanline_width];
		  rgbe2float(&data[RGBE_DATA_RED],&data[RGBE_DATA_GREEN],
			 &data[RGBE_DATA_BLUE],rgbe);
		  data += RGBE_DATA_SIZE;
		}
		num_scanlines--;
  }
  free(scanline_buffer);
  return RGBE_RETURN_SUCCESS;
}